/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Activate/Deactivate code for hierarchical RAID Sets.
 */

#include <stdlib.h>
#include "internal.h"
#include "devmapper.h"

/* Make a dm path. */
char *mkdm_path(const char *name)
{
	char *ret;
	const char *dir = get_dm_dir();

	if (!(ret = dbg_malloc(strlen(dir) + strlen(name) + 2)))
		LOG_ERR(NULL, "Failed allocating dm name");

	sprintf(ret, "%s/%s", dir, name);
	return ret;
}

/*
 * Levels definitions.
 *
 * Must be in the same order as in the level_type enum !!!
 */
const char *levels[] = {
	NULL,
	"partition",
	"spare",
	"linear",
	"striped",
	"mirror",
	/* RAID levels below not supported (yet) */
};

/*
 * Status definitions.
 *
 * Must be in the same order as in the status_type enum !!!
 */
const char *status[] = {
	NULL,
	"setup",
	"operational",
	"broken",
	"nosync",
};

/* Fetch the respective ASCII string off the levels array. */
const char *get_level(enum level_type l)
{
	return l >= ARRAY_SIZE(levels) ? NULL : levels[l];
}

/* Return the ASCII level for a RAID set. */
const char *get_set_level(void *rs)
{
	return get_level(((struct dmraid_set*) rs)->level);
}

/* Fetch the respective ASCII string off the levels array. */
const char *get_status(enum status_type s)
{
	return s >= ARRAY_SIZE(status) ? NULL : status[s];
}

/* Grow a string. */
static int grow_string(char **string, size_t len)
{
	char *tmp = *string;

 	len += tmp ? strlen(tmp) + 1 : 1;
	*string = dbg_realloc(*string, len);
	if (*string) {
		if (!tmp)
			**string = '\0';
	} else if (tmp)
		dbg_free(tmp);
	
	return *string ? 1 : 0;
}

/* Free a string. */
static void free_string(char **string)
{
	if (*string) {
		dbg_free(*string);
		*string = NULL;
	}
}

/* Push a string onto the end of another. */
static int p_str(char **string, const char *s)
{
	if (!grow_string(string, strlen(s)))
		return 0;

	strcat (*string, s);
	return 1;
}

/* Push an uint64_t in ascii onto the end of a string. */
static int p_u64(char **string, uint64_t u)
{
	char buffer[32];

	sprintf(buffer, "%" PRIu64, u);
	if (!grow_string(string, strlen(buffer)))
		return 0;

	sprintf(&(*string)[strlen(*string)], "%" PRIu64, u);
	return 1;
}

/* Find smallest disk size above given minimum. */
static uint64_t _smallest(struct dmraid_set *rs, uint64_t min)
{
	uint64_t ret = ~0, sectors;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	if (min % rs->stride)
		min += rs->stride;

	list_for_each_entry(r, &rs->sets, next) {
		sectors = calc_total_sectors(r);
		if (sectors > min &&
		    (sectors < ret))
			ret = sectors;
	}

	list_for_each_entry(rd, &rs->devs, next) {
		if (rd->sectors > min &&
		    rd->sectors < ret)
			ret = rd->sectors;
	}

	return ret & ~(((uint64_t) rs->stride) >> 1);
}

/*
 * Definitions of mappings.
 */

/* Undefined mapping. */
static int dm_undef(struct dmraid_set *rs, char **table,
		    const char *level_ascii)
{
	LOG_ERR(0, "Undefined RAID level on %s", rs->name);
}

/* "Spare mapping". */
static int dm_spare(struct dmraid_set *rs, char **table,
		    const char *level_ascii)
{
	log_verbose(1, "Spare device %s", rs->name);
	return 1;
}

/*
 * Create dm table for linear mapping.
 */
static void _dm_linear(char **table, const char *prefix, const char *path,
		       const char *level_ascii,
		       uint64_t start, uint64_t sectors, uint64_t offset)
{
	p_u64(table, start);
	p_str(table, " ");
	p_u64(table, sectors);
	p_str(table, " ");
	p_str(table, level_ascii);
	p_str(table, " ");
	if (prefix) {
		p_str(table, prefix);
		if (*table[strlen(*table)-1] != '/')
			p_str(table, "/");
	}

	p_str(table, path);
	p_str(table, " ");
	p_u64(table, offset);
}

static int dm_linear(struct dmraid_set *rs, char **table,
		     char const *level_ascii)
{
	uint64_t start = 0, sectors = 0;
	struct dmraid_dev *rd;
	struct dmraid_set *r;

	list_for_each_entry(r, &rs->sets, next) {
		sectors = calc_total_sectors(r);
		_dm_linear(table, get_dm_dir(), r->name, level_ascii,
			   start, sectors, 0);
		start += sectors;
		if (!list_empty(&r->next))
			p_str(table, "\n");
	}

	list_for_each_entry(rd, &rs->devs, next) {
		_dm_linear(table, NULL, rd->di->path, level_ascii,
			   start, rd->sectors, rd->offset);
		start += rd->sectors;
		if (rd->next.next != &rs->devs)
			p_str(table, "\n");
	}

	return 1;
}

/*
 * Create dm table for a partitions mapping.
 *
 * Operates on a partitioned RAID set with
 * 1 RAID device defining a partition mapping.
 */
static int dm_partitions(struct dmraid_set *rs, char **table,
			 const char *level_ascii)
{
	return dm_linear(rs, table, get_level(l_linear));
}

/*
 * Create dm table for striped mapping taking
 * different disk sizes and the stride size into acccount.
*/
static void _dm_path_offset(char **table, const char *path, uint64_t offset)
{
	p_str(table, " ");
	p_str(table, path);
	p_str(table, " ");
	p_u64(table, offset);
}

static void _dm_striped_start(char **table, uint64_t min, uint64_t last_min,
			      unsigned int n, unsigned int stride)
{
	p_u64(table, n * last_min);
	p_str(table, " ");
	p_u64(table, n * (min - last_min));
	p_str(table, " striped ");
	p_u64(table, n);
	p_str(table, " ");
	p_u64(table, stride);
}

static int dm_striped(struct dmraid_set *rs, char **table,
		      char const *level_ascii)
{
	unsigned int dsks = count_devs(rs), n;
	uint64_t min = 0, last_min;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	while (dsks) {
		last_min = min;
		if (!(min = _smallest(rs, min)))
			return 0;
	
		n = 0;
		list_for_each_entry(r, &rs->sets, next) {
			if (calc_total_sectors(r) > last_min)
				n++;
		}
	
		list_for_each_entry(rd, &rs->devs, next) {
			if (rd->sectors > last_min)
				n++;
		}
	
		_dm_striped_start(table, min, last_min, n, rs->stride);
		list_for_each_entry(r, &rs->sets, next) {
			if (calc_total_sectors(r) > last_min) {
				char *path = mkdm_path(r->name);

				_dm_path_offset(table, path, 0);
				dbg_free(path);
				dsks--;
			}
		}
	
		list_for_each_entry(rd, &rs->devs, next) {
			if (rd->sectors > last_min) {
				_dm_path_offset(table, rd->di->path,
						rd->offset);
				dsks--;
			}
		}
	
		if (dsks)
			p_str(table, "\n");
	}

	return 1;
}

/*
 * Create dm table for mirrored mapping.
 */
#define ONE_MB	2*1024*1024
static int dm_raid1(struct dmraid_set *rs, char **table,
		    char const *level_ascii)
{
	uint64_t sectors = 0;
	unsigned int max, mirrors = 0, region_size;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, next)
		mirrors++;

	list_for_each_entry(rd, &rs->devs, next)
		mirrors++;

	if (!(sectors = _smallest(rs, 0)))
		return 0;

	/* Calculate dirty log region size */
	max = sectors / 1024;
	if (max > 64 * ONE_MB)
		max = 64 * ONE_MB;
	for (region_size = 128; region_size < max; region_size <<= 1);
	region_size >>= 1;

	p_str(table, "0 ");
	p_u64(table, sectors);
	p_str(table, " ");
	p_str(table, level_ascii);
	p_str(table, " core 2 ");
	p_u64(table, region_size);
	p_str(table, rs->status == s_broken ? " sync " : " nosync ");
	p_u64(table, mirrors);

	list_for_each_entry(r, &rs->sets, next)
		_dm_path_offset(table, mkdm_path(r->name), 0);

	list_for_each_entry(rd, &rs->devs, next)
		_dm_path_offset(table, rd->di->path, rd->offset);

	return 1;
}

/*
 * Activate/deactivate of (sub)sets.
 */

/*
 * Array of handler functions for the various level types.
 *
 * Functions must be in the same order as in the level_type enum !!!
 *
 */
struct level_function {
	int(*f)(struct dmraid_set *rs, char **table, char const *level_ascii);
} level_functions[] = {
	{ dm_undef },
	{ dm_partitions },
	{ dm_spare },
	{ dm_linear },
	{ dm_striped },
	{ dm_raid1 },
	/* RAID levels below not supported (yet) */
};

/* Deactivate a single set (one level of a device stack). */
static int deactivate_superset(struct dmraid_set *rs)
{
	int ret = 1;

	if (test())
		log_print("%s", rs->name);
	else if ((ret = dm_status(rs)))
		ret = dm_remove(rs);
	else
		log_err("RAID set \"%s\" is not active", rs->name);

	return ret;
}

/* Deactivate a RAID set. */
int deactivate_set(void *v)
{
	struct dmraid_set *r, *rs = v;

	/* Deactivate myself */
	if (!deactivate_superset(rs))
		return 0;

	/* Deactivate any subsets recursively */
	list_for_each_entry(r, &rs->sets, next)
		deactivate_set((void *) r);

	return 1;
}

/* Activate a single set. */
static int activate_subset(struct dmraid_set *rs)
{
	int ret = 0;
	char const *level_ascii;
	char *table = NULL;

	if (!(level_ascii = get_level(rs->level)))
		LOG_ERR(0, "Failed to recognize device-mapper mapping type");

	/* Call type handler */
	if ((ret = (&level_functions[rs->level])->f(rs, &table, level_ascii))) {
		if (test()) {
			log_print("Mapping table for \"%s\":", rs->name);
			log_print("%s", table);
		} else if (dm_status(rs)) {
			log_err("%s already active!", rs->name);
			ret = 0;
		} else
			ret = dm_create(rs, table);
	}

	free_string(&table);

	return ret;
}

/* Activate a RAID set recursively (eg, RAID1 on top of RAID0). */
int activate_set(void *v)
{
	int ret = 0;
	struct dmraid_set *r, *rs = v;

	if (dm_status(rs)) {
		log_print("RAID set \"%s\" already active", rs->name);
		return 0;
	}

	/* Recursively walk down the chain of stacked RAID sets */
	list_for_each_entry(r, &rs->sets, next) {
		/* Activate set below this one */
		if (!(ret = activate_set((void *) r)))
			goto out;
	}

	if (!(ret = activate_subset(rs)))
		deactivate_set(rs);

  out:
	return ret;
}
