/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"

/* Display information about a block device */
static void log_device(struct disk_info *di)
{
	log_print(column() ? "%s,%u,%u,%u,%" PRIu64 :
			     "%s: %4u heads, %3u sectors, %6u cylinders, "
			     "%12" PRIu64 " sectors total",
		  di->path, di->geometry.heads,
		  di->geometry.sectors, di->geometry.cylinders,
		  di->sectors);
}

/* Display information about a RAID device */
static void log_rd(struct dmraid_dev *rd)
{
	if (!rd)
		return;

	if (rd->sectors) {
		const char *type = rd->fmt->type ?
				   get_level(rd->fmt->type(rd)) : "unsupported";

		if (!type)
			type = "unknown";

		log_print(column() ? "%s,%s,%s,%s,%" PRIu64 ",%" PRIu64 :
				     "%s: %s, \"%s\", %7s, %" PRIu64
				     " sectors, data@ %" PRIu64,
			  rd->di->path, rd->name, rd->fmt->name, type,
		 	  rd->sectors, rd->offset);
	} else log_print(column() ? "%s,%s,free" :
				     "%s: \"%s\" format, not used",
			 rd->di->path, rd->fmt->name);
}

/* Log native RAID device information. */
static void log_rd_native(struct dmraid_dev *rd)
{
	if (!rd)
		return;

	if (rd->fmt->log) {
		rd->fmt->log(rd);
		log_print(" ");
	} else
		log_print("\"%s\" doesn't support logging of RAID "
			  "device information", rd->fmt->name);
}


/* Display information about a dmraid format handler */
static void log_format(struct dmraid_format *fmt)
{
	log_print("%-6s : %s", fmt->name, fmt->descr);
}

/* Display information about found block devices */
void display_devices(enum dev_type type)
{
	int devs;
	void *iter;
	struct disk_info *di;

	if ((devs = count_devices(type)))
		log_verbose(1, "%s device%s discovered:\n",
			    type == RAID ? "RAID" : "Block",
			    devs == 1 ? "" : "s");

	for (iter = rewind_disk_info(); (di = get_disk_info(&iter));) {
		if (type == RAID)
			log_rd(di->rd);
		else if (type == NATIVE)
			log_rd_native(di->rd);
		else
			log_device(di);
	}

	if (!devs)
		log_print("No Software RAID devices found\n");
}

static void __display_set(struct dmraid_set *rs)
{
	uint64_t sectors;
	struct dmraid_dev *rd;
	const char *type   = get_level(rs->level);
	const char *status = get_status(rs->status);

	if (!type)
		type = "unknown";

	if (!status)
		status = "unknown";

	sectors = calc_total_sectors(rs);
	if (rs->level == l_raid1 && !list_empty(&rs->devs)) {
		rd = list_entry(rs->devs.next, struct dmraid_dev, next);
		sectors = rd->sectors;
	}

	if (column())
		log_print("%s,%" PRIu64 ",%u,%s,%s,%u",
			  rs->name, sectors, rs->stride,
			  type, status, count_devs(rs));
	else {
		log_print("name  : %s", rs->name);
		log_print("size  : %" PRIu64, sectors);
		if (rs->stride)
			log_print("stride: %u", rs->stride);
		log_print("type  : %s", type);
		log_print("status: %s", status);
		log_print("dev   : %d", count_devs(rs));
	}

	list_for_each_entry(rd, &rs->devs, next)
		log_rd(rd);
}

/* FIXME: Spock, do something better (neater). */
static void _display_set(struct dmraid_set *rs, int top)
{
	struct dmraid_set *r;

	if (!column())
		log_print("%s %s%s%set",
		  	top ? "--->" : "\n***",
		  	rs->status == s_inconsistent ? "*Inconsistent* " : "",
		  	dm_status(rs) ? "Active " : "",
		  	!list_empty(&rs->sets) ? "Supers" : (top ? "Subs" : "S"));

	__display_set(rs);
	list_for_each_entry(r, &rs->sets, next)
		_display_set(r, top + 1);
}

void display_set(void *v, enum active_type active)
{
	struct dmraid_set *rs = v;
	int dmstatus = dm_status(rs);

	if (active == ALL ||
	    (active == A && dmstatus) ||
	    (active == I && !dmstatus))
		_display_set(rs, 0);
}

/*
 * Display information about supported RAID metadata formats
 * (ie. registered format handlers)
 */
int list_formats(void)
{
	void *iter;
	struct dmraid_format *fmt;

	iter = get_format_list();
	if ((fmt = get_format(&iter, FMT_ALL)))
		log_verbose(1, "Supported Software RAID sets:\n");

	do {
		log_format(fmt);
	} while ((fmt = get_format(&iter, FMT_ALL)));

	log_print("(0): Recognize, (+): Recognize+Activate");

	return 1;
}
