/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Highpoint 37X ATARAID format handler
 *
 * hpt37x_read(), hpt37x_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "hpt37x.h"
#include <utils.h>

static const char *handler = "hpt37x";

/* Make up RAID set name from magic_[01] numbers */
/* FIXME: better name ? */
static size_t __name(struct hpt37x *hpt, char *str, size_t len,
		     unsigned char type)
{
	const char *fmt = (hpt->magic_1 && type) ? "hpt37x_%u-%u" : "hpt37x_%u";
		
	return snprintf(str, len, fmt,
			hpt->magic_0, hpt->magic_1) + 1;
}

static char *_name(struct hpt37x *hpt, unsigned char type) {
	size_t len;
	char *ret;

	len  = __name(hpt, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(hpt, ret, len, type);
		mk_alpha(ret + 7);
	} else
		log_err("%s: Failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status_type _status(struct dmraid_dev *rd)
{
	struct hpt37x* hpt = (struct hpt37x*) rd->meta_areas[0].area;

	return hpt->magic == HPT37X_MAGIC_BAD ? s_broken : s_operational;
}

/* Mapping of HPT 37X types to generic types. */
static struct levels levels [] = {
	{ HPT37X_T_SINGLEDISK, l_linear},
	{ HPT37X_T_SPAN, l_linear},
	{ HPT37X_T_RAID_0, l_raid0},
	{ HPT37X_T_RAID_1, l_raid1},
	{ HPT37X_T_RAID_01_RAID_0, l_raid1},
	/* FIXME: support RAID 3+5 */
	{ 0, l_undef}
};

/* Neutralize disk type. */
static enum level_type hpt37x_type(struct dmraid_dev *rd)
{
	return rd_type(levels, (unsigned int) ((struct hpt37x*) rd->meta_areas[0].area)->type);
}

/*
 * Decide about ordering sequence of RAID device.
 * Called by list_add_dev_sorted().
 */
static int _hpt37x_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct hpt37x*) new->meta_areas[0].area)->disk_number <
	       ((struct hpt37x*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Read a Highpoint 37X RAID device.
 */
/* Endianess conversion. */
static void to_cpu(struct hpt37x *hpt)
{
	struct hpt37x_errorlog *log;

	hpt->magic = le32_to_cpu(hpt->magic);
	hpt->magic_0 = le32_to_cpu(hpt->magic_0);
	hpt->magic_1 = le32_to_cpu(hpt->magic_1);
	hpt->order = le32_to_cpu(hpt->order);
	hpt->total_secs = le32_to_cpu(hpt->total_secs);
	hpt->disk_mode = le32_to_cpu(hpt->disk_mode);
	hpt->boot_mode = le32_to_cpu(hpt->boot_mode);

	for (log = &hpt->errorlog[0];
	     log < &hpt->errorlog[hpt->error_log_entries];
	     log++) {
		log->timestamp = le32_to_cpu(log->timestamp);
		log->lba = le32_to_cpu(log->lba);
	}
}

/* Magic check. */
static int hpt37x(struct hpt37x *hpt)
{
	return (hpt->magic == HPT37X_MAGIC_OK ||
		hpt->magic == HPT37X_MAGIC_BAD) &&
		hpt->disk_number < 8;
}

static int setup_rd(struct dmraid_dev *rd, struct hpt37x *hpt,
		    struct disk_info *di);
static struct dmraid_dev *hpt37x_read(struct disk_info *di) {
	struct hpt37x *hpt;
	struct dmraid_dev *rd = NULL;

	if (!(hpt = alloc_private(handler, sizeof(*hpt))))
		goto out;

	if (!read_dev(di->path, (void *) hpt, HPT37X_CONFIGOFFSET,
		      sizeof(*hpt)))
		goto bad;

	to_cpu(hpt);
	if (!hpt37x(hpt))
		goto bad;

	file_metadata(handler, di, hpt, sizeof(*hpt));

	if (!(rd = alloc_dev()))
		goto bad;

	if (setup_rd(rd, hpt, di))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(hpt);
   out:
	return rd;
}

/*
 * Write a Highpoint 37X RAID device.
 */
static int hpt37x_write(struct dmraid_dev *rd, int erase)
{
	int ret;
	struct hpt37x *hpt = rd->meta_areas[0].area;

	to_disk(hpt);
	ret = write_metadata(handler, rd, -1, erase);
	to_cpu(hpt);
	return ret;
}

/*
 * Group the RAID disk into a set.
 *
 * Check device hierarchy and create sub sets appropriately.
 *
 */
static int _alloc_subsets(struct dmraid_set *rs,
			  struct dmraid_dev *rd,
			  struct hpt37x *hpt)
{
	unsigned int i;
	struct dmraid_set *r;

	/* Allocate 2 subsets and link them to the superset */
	for (i = 0; i < 2; i++) {
		if ((r = alloc_set())) {
			r->level = (hpt->type == HPT37X_T_RAID_01_RAID_0) ? l_raid0 : l_raid1;
			list_add_tail(&r->next, &rs->sets);
		} else {
			if (i) {
				free_set(list_entry(rs->sets.next, typeof(*r), next));
				INIT_LIST_HEAD(&rs->sets);
			}

			return 0;
		}
	}

	/* Set properties of superset */
	rs->name = _name(hpt, 0);
	rs->level = (hpt->type == HPT37X_T_RAID_01_RAID_0) ? l_raid1 : l_raid0;
	rs->status = _status(rd);
	return 1;
}

static int group_rd(struct dmraid_set *rs, struct dmraid_dev *rd)
{
	unsigned int sorted;
	struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area, *h = NULL;
	struct dmraid_set *r;

	if (rd_type(levels, hpt->type) == l_undef)
		LOG_ERR(0, "hpt37x - RAID type %u not supported", hpt->type);

	if (!list_empty(&rs->devs))
		h = (struct hpt37x*) list_entry(rs->devs.next, typeof(*rd), next)->meta_areas[0].area;

	rs->stride = 1 << hpt->raid0_shift;

	switch (hpt->type) {
	case HPT37X_T_RAID_01_RAID_0:
	case HPT37X_T_RAID_01_RAID_1:
		if (h && h->magic_1 != hpt->magic_1)
			LOG_ERR(0, "hpt37x: magic_1 mismatch");

		/* Allocate 2 subsets for the stripe/mirror on top */
		if (list_empty(&rs->sets)) {
			if (!_alloc_subsets(rs, rd, hpt))
				return 0;
		}

		/* Sort RAID disk into appropriate subset */
		sorted = 0;
		list_for_each_entry(r, &rs->sets, next) {
			if (list_empty(&r->devs))
				rs->name = rd->name;
			else
				h = (struct hpt37x*) list_entry(r->devs.next, typeof(*rd), next)->meta_areas[0].area;

			if (list_empty(&r->devs) ||
			    h->magic_0 == hpt->magic_0) {
				list_add_dev_sorted(r, rd, _hpt37x_sort);
				sorted++;
				break;
			}
		}

		if (!sorted)
			goto out;

		break;

	default:
		if (h && h->magic_0 != hpt->magic_0)
			goto out;

		list_add_dev_sorted(rs, rd, _hpt37x_sort);
		break;
	}

	return 1;

  out:
	log_err("%s: magic_0 mismatch", handler);
	return 0;
}

/*
 * Add a Highpoint RAID device to a set.
 */
static struct dmraid_set *hpt37x_group(struct dmraid_dev *rd) {
	struct dmraid_set *rs;

	if (!(rs = find_set(rd->name, 0))) {
		if (!(rs = alloc_set()))
			return NULL;

		rs->name = rd->name;
		rs->level = hpt37x_type(rd);
		rs->status = _status(rd);
		list_add_tail(&rs->next, default_set_list());
	}

	return group_rd(rs, rd) ? rs : NULL;
}

/*
 * Check a Highpoint RAID set.
 *
 * FIXME: more sanity checks.
 */
static int hpt37x_check(struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;
	struct dmraid_set *r;

	list_for_each_entry(r, &rs->sets, next) {
		if (!hpt37x_check(r))
			return 0;
	}

	devs = count_devs(rs);
	list_for_each_entry(rd, &rs->devs, next) {
		struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area;

		if (devs != hpt->raid_disks) {
			rs->status = s_inconsistent;
			LOG_ERR(0, "hpt37x: ERROR, wrong # of devices in RAID "
				   "set \"%s\" [%u/%u]",
				   rs->name, devs, hpt->raid_disks);
		}
	}

	return 1;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an HPT37X RAID device.
 */
static void hpt37x_log(struct dmraid_dev *rd)
{
	unsigned int i;
	struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area;

	log_print("%s (%s):", rd->di->path, handler);
	log_print("magic: 0x%x", hpt->magic);
	log_print("magic_0: 0x%x", hpt->magic_0);
	log_print("magic_1: 0x%x", hpt->magic_1);
	log_print("order: %u", hpt->order);
	log_print("raid_disks: %u", hpt->raid_disks);
	log_print("raid0_shift: %u", hpt->raid0_shift);
	log_print("type: %u", hpt->type);
	log_print("disk_number: %u", hpt->disk_number);
	log_print("total_secs: %u", hpt->total_secs);
	log_print("disk_mode: 0x%x", hpt->disk_mode);
	log_print("boot_mode: 0x%x", hpt->boot_mode);
	log_print("boot_disk: %u", hpt->boot_disk);
	log_print("boot_protect: %u", hpt->boot_protect);
	log_print("error_log_entries: %u", hpt->error_log_entries);
	log_print("error_log_index: %u", hpt->error_log_index);
	if (hpt->error_log_entries)
		log_print("error_log:");

	for (i = 0; i < 32; i++) {
		if (!hpt->errorlog[i].timestamp)
			break;

		log_print("timestamp: %u", hpt->errorlog[i].timestamp);
		log_print("reason: %u", hpt->errorlog[i].reason);
		log_print("disk: %u", hpt->errorlog[i].disk);
		log_print("status: %u", hpt->errorlog[i].status);
		log_print("sectors: %u", hpt->errorlog[i].sectors);
		log_print("lba: %u", hpt->errorlog[i].lba);
       };
}
#endif

static struct dmraid_format hpt37x_format = {
	.name	= "hpt37x",
	.descr	= "(+) Highpoint HPT37X",
	.fmt_type = FMT_RAID,
	.read	= hpt37x_read,
	.write	= hpt37x_write,
	.group	= hpt37x_group,
	.check	= hpt37x_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= hpt37x_log,
#endif
	.type	= hpt37x_type,
};

/* Register this format handler with the format core */
int register_hpt37x(void) {
	register_format(&hpt37x_format);
	return 1;
}

/* Derive the RAID device contents from the Highpoint ones. */
static int setup_rd(struct dmraid_dev *rd, struct hpt37x *hpt,
		    struct disk_info *di)
{
	rd->di = di;
        rd->name = _name(hpt, 1);
	rd->fmt = &hpt37x_format;

        rd->offset = HPT37X_DATAOFFSET; /* Data offset from start of device */
        rd->sectors = hpt->total_secs;
        rd->sectors += rd->sectors & 1;
        rd->sectors /= hpt->raid_disks;

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = HPT37X_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*hpt);
	rd->meta_areas[0].area = (void*) hpt;

	return 1;
}
