/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Highpoint 45X ATARAID format handler
 *
 * hpt45x_read(), hpt45x_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "hpt45x.h"
#include <utils.h>

static const char *handler = "hpt45x";

/* Make up RAID set name from magic_0 number */
/* FIXME: better name ? */
static size_t __name(struct hpt45x *hpt, char *str, size_t len,
		     unsigned char type)
{
	const char *fmt = type ? "hpt45x_%u-%u" : "hpt45x_%u";

	return snprintf(str, len, fmt, hpt->magic_0,
			hpt->raid1_disk_number) + 1;
}

static char *_name(struct hpt45x *hpt, unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(hpt, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(hpt, ret, len, type);
		mk_alpha(ret + 7);
	} else
		log_err("%s: Failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status_type _status(struct dmraid_dev *rd)
{
	struct hpt45x* hpt = (struct hpt45x*) rd->meta_areas[0].area;

	return hpt->magic == HPT45X_MAGIC_BAD ? s_broken : s_operational;
}

/* Mapping of HPT 45X types to generic types */
static struct levels levels [] = {
	{ HPT45X_T_SPAN, l_linear},
	{ HPT45X_T_RAID_0, l_raid0},
	{ HPT45X_T_RAID_1, l_raid1},
/* FIXME: handle RAID 3+5 */
	{ 0, l_undef}
};

/* Neutralize disk type */
static enum level_type hpt45x_type(struct dmraid_dev *rd) {
	return rd_type(levels,
		       (unsigned int) ((struct hpt45x*) rd->meta_areas[0].area)->type);
}

/*
 * Decide about ordering sequence of RAID device.
 * Called by list_add_dev_sorted().
 */
static int _hpt45x_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct hpt45x*) new->meta_areas[0].area)->disk_number <
	       ((struct hpt45x*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Group the RAID disk into a HPT45X set.
 *
 * Check device hierarchy and create super set appropriately.
 */
static int group_rd(struct dmraid_set *rs, struct dmraid_dev *rd)
{
	unsigned int sorted;
	char *name;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;
	struct hpt45x *h = NULL;
	struct dmraid_set *r, *super;

	if (rd_type(levels, hpt->type) == l_undef)
		LOG_ERR(0, "hpt37x - RAID type %u not supported", hpt->type);

	if (!list_empty(&rs->devs))
		h = (struct hpt45x*) list_entry(rs->devs.next, typeof(*rd), next)->meta_areas[0].area;

	switch (hpt->type) {
	case HPT45X_T_RAID_0:
		if (hpt->raid1_type != HPT45X_T_RAID_1)
			goto no_raid10;
		/*
		 * We've got a striped raid set with a mirror on top
		 * when we get here.
		 * Let's find and optionally allocate the mirror set on top.
		 */

		/* Superset name */
		if (!(name = _name(hpt, 0)))
			return 0;

		if ((super = find_set(name, 0)))
			dbg_free(name);
		else {
			if (!(super = alloc_set())) {
				dbg_free(name);
				return 0;
			}

			super->name = name;
			super->level = l_raid1;
			super->status = _status(rd);
			super->stride = 1 << hpt->raid1_shift;
			list_add_tail(&super->next, default_set_list());
		}

		rs->stride = 1 << hpt->raid0_shift;
		if (!find_set(rs->name, 1))
			list_add_tail(&rs->next, &super->sets);

		/* Sort RAID disk into appropriate subset */
		sorted = 0;
		list_for_each_entry(r, &super->sets, next) {
			if (!list_empty(&r->devs))
				h = (struct hpt45x*) list_entry(r->devs.next, typeof(*rd), next)->meta_areas[0].area;

			if (list_empty(&r->devs) ||
			    (h->magic_0 == hpt->magic_0 &&
			     h->raid1_disk_number == hpt->raid1_disk_number)) {
				list_add_dev_sorted(r, rd, _hpt45x_sort);
				sorted++;
				break;
			}
		}

		if (!sorted)
			goto out;

		break;

	case HPT45X_T_SPAN:
	case HPT45X_T_RAID_1:
  no_raid10:
		if (!find_set(rs->name, 0))
			list_add_tail(&rs->next, default_set_list());

		if (!h || (h->magic_0 == hpt->magic_0)) {
			rs->stride = 1 << hpt->raid0_shift;
			list_add_dev_sorted(rs, rd, _hpt45x_sort);
		} else
			goto out;

		break;

	default:
		LOG_ERR(0, "hpt - RAID type %d not supported on %s",
			hpt->type, rd->di->path);
	}

	return 1;

  out:
	log_err("%s: magic_0 mismatch for %s", handler, rd->di->path);
	return 0;
}

/*
 * Add a Highpoint RAID device to a set.
 */
static struct dmraid_set *hpt45x_group(struct dmraid_dev *rd)
{
	struct dmraid_set *rs;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

	/* Unused disk */
	if (!hpt->magic_0)
		return 0;

	if (!(rs = find_set(rd->name, 1))) {
		if (!(rs = alloc_set()))
			return NULL;

		rs->name   = rd->name;
		rs->level   = hpt45x_type(rd);
		rs->status = _status(rd);
	}

	return group_rd(rs, rd) ? rs : NULL;
}

/*
 * Read a Highpoint 45X RAID device.
 */
/* Endianess conversion. */
static void to_cpu(struct hpt45x *hpt)
{
	hpt->magic = le32_to_cpu(hpt->magic);
	hpt->magic_0 = le32_to_cpu(hpt->magic_0);
	hpt->magic_1 = le32_to_cpu(hpt->magic_1);
	hpt->total_secs = le32_to_cpu(hpt->total_secs);
}

/* Magic check. */
static int hpt45x(struct hpt45x *hpt)
{
	return (hpt->magic == HPT45X_MAGIC_OK ||
		hpt->magic == HPT45X_MAGIC_BAD) &&
		hpt->disk_number < 8;
}

static int setup_rd(struct dmraid_dev *rd, struct hpt45x *hpt,
		    struct disk_info *di);
static struct dmraid_dev *hpt45x_read(struct disk_info *di)
{
	struct hpt45x *hpt;
	struct dmraid_dev *rd = NULL;

	if (!(hpt = alloc_private(handler, sizeof(*hpt))))
		goto out;

	if (!read_dev(di->path, (void *) hpt,
		      HPT45X_CONFIGOFFSET, sizeof(*hpt)))
		goto bad;

	to_cpu(hpt);
	if (!hpt45x(hpt))
		goto bad;

	file_metadata(handler, di, hpt, sizeof(*hpt));

	if (!(rd = alloc_dev()))
		goto bad;

	if (setup_rd(rd, hpt, di))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(hpt);

   out:
	return rd;
}

/*
 * Write a Highpoint 45X RAID device.
 */
static int hpt45x_write(struct dmraid_dev *rd, int erase)
{
	int ret;
	struct hpt45x *hpt = rd->meta_areas[0].area;

	to_disk(hpt);
	ret = write_metadata(handler, rd, -1, erase);
	to_cpu(hpt);
	return ret;
}

/*
 * Check integrity of a Highpoint 45X RAID set.
 *
 * FIXME: more sanity checks.
 */
static int hpt45x_check(struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;
	struct dmraid_set *r;

	list_for_each_entry(r, &rs->sets, next) {
		if (!hpt45x_check(r))
			return 0;
	}

	devs = count_devs(rs);
	list_for_each_entry(rd, &rs->devs, next) {
		struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

		if (devs != hpt->raid_disks) {
			rs->status = s_inconsistent;
			LOG_ERR(0, "%s: ERROR, wrong # of devices in RAID "
			   	"set \"%s\" [%u/%u]",
			   	handler, rs->name, devs, hpt->raid_disks);
		}
	}

	return 1;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an HPT45X RAID device.
 */
static void hpt45x_log(struct dmraid_dev *rd)
{
	unsigned int i;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

	log_print("%s (%s):", rd->di->path, handler);
	log_print("magic: 0x%x", hpt->magic);
	log_print("magic_0: 0x%x", hpt->magic_0);
	log_print("magic_1: 0x%x", hpt->magic_1);
	log_print("total_secs: %u", hpt->total_secs);
	log_print("type: %u", hpt->type);
	log_print("raid_disks: %u", hpt->raid_disks);
	log_print("disk_number: %u", hpt->disk_number);
	log_print("raid0_shift: %u", hpt->raid0_shift);
	for (i = 0; i < 3; i++)
		log_print("dummy[i]: 0x%x", hpt->dummy[i]);
	log_print("raid1_type: %u", hpt->raid1_type);
	log_print("raid1_raid_disks: %u", hpt->raid1_raid_disks);
	log_print("raid1_disk_number: %u", hpt->raid1_disk_number);
	log_print("raid1_shift: %u", hpt->raid1_shift);
	for (i = 0; i < 3; i++)
		log_print("dummy1[i]: 0x%x", hpt->dummy1[i]);
}
#endif

static struct dmraid_format hpt45x_format = {
	.name	= "hpt45x",
	.descr	= "(+) Highpoint HPT45X",
	.fmt_type = FMT_RAID,
	.read	= hpt45x_read,
	.write	= hpt45x_write,
	.group	= hpt45x_group,
	.check	= hpt45x_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= hpt45x_log,
#endif
	.type	= hpt45x_type,
};

/* Register this format handler with the format core */
int register_hpt45x(void)
{
	register_format(&hpt45x_format);
	return 1;
}

/* Set the RAID device contents up derived from the Highpoint ones */
static int setup_rd(struct dmraid_dev *rd, struct hpt45x *hpt,
		    struct disk_info *di)
{
	rd->di = di;
	rd->name = _name(hpt, hpt->raid1_type);
	rd->fmt = &hpt45x_format;

	rd->offset = HPT45X_DATAOFFSET;
	rd->sectors = hpt->total_secs;
	rd->sectors += rd->sectors & 1; /* Round up to KBs */
 	rd->sectors /= hpt->raid_disks;

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = HPT45X_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*hpt);
	rd->meta_areas[0].area = (void*) hpt;

	return 1;
}
