/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Intel ATARAID metadata format handler.
 *
 * isw_read() profited from Carl-Daniel Hailfinger's raiddetect code.
 *
 * Needs more metadata reenginieering and grouping logic coding.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "isw.h"
#include <utils.h>

static const char *handler = "isw";

/*
 * Make up RAID set name from familyNum.
 *
 * FIXME: better name rather than deriving from first time checksum ?
 */
static size_t __name(struct isw *isw, char *str, size_t len)
{
	return snprintf(str, len, "isw_%u", isw->familyNum) + 1;
}

static char *_name(struct isw *isw)
{
        size_t len;
        char *ret;

        len = __name(isw, NULL, 0);
        if ((ret = dbg_malloc(len + 1))) {
                __name(isw, ret, len);
		mk_alpha(ret);
        } else
                log_err("%s: Failed allocating name string", handler);

        return ret;
}

/*
 * Mapping of Intel types to generic types.
 *
 * FIXME: Code ;)
 */
static struct levels levels[] = {
        { 0, l_undef}
};

/* Neutralize disk type. */
static enum level_type isw_type(struct dmraid_dev *rd)
{
	return levels[0].type;
}

/*
 * Generate checksum of Raid metadata for mpbSize/sizeof(u32) words
 * (checksum field itself ignored for this calculation).
 */
static uint32_t _checksum(struct isw *isw)
{
	uint32_t end = isw->mpbSize / sizeof(end), ret = 0;
	uint32_t *p = (uint32_t*) isw;

	while (end--)
		ret += *p++;

	return ret - isw->checkSum;
}

/*
 * Read an Intel RAID device
 */
/* Endianess conversion. */
static void to_cpu(struct isw *isw, int full)
{
	struct isw_diskTbl *tbl;
	struct isw_RaidDev* ird, *end;

	isw->checkSum = le32_to_cpu(isw->checkSum);
	isw->mpbSize = le32_to_cpu(isw->mpbSize);
	isw->familyNum = le32_to_cpu(isw->familyNum);
	isw->generationNum = le32_to_cpu(isw->generationNum);

	if (!full)
		return;

	for (tbl = &isw->diskTbl[0];
	     tbl < &isw->diskTbl[isw->numDisks];
	     tbl++) {
		tbl->totalBlocks = le32_to_cpu(tbl->totalBlocks);
		tbl->scsiId = le32_to_cpu(tbl->scsiId);
		tbl->status = le32_to_cpu(tbl->status);
	}

	for (ird = raiddev(isw), end = &ird[isw->numRaidDevs];
	     ird < end;
	     ird++) {
		ird->totalSize = le32_to_cpu(ird->totalSize);
		ird->thisDevSize = le32_to_cpu(ird->thisDevSize);
		ird->thisDevSize256 = le32_to_cpu(ird->thisDevSize256);
		ird->stride = le32_to_cpu(ird->stride);
	}
}

static int setup_rd(struct dmraid_dev *rd, struct isw *isw,
		    struct disk_info *di, loff_t isw_sboffset,
		    size_t meta_size);
static struct dmraid_dev *isw_read(struct disk_info *di)
{
	size_t size;
	loff_t mpbblocks, isw_sboffset = ISW_CONFIGOFFSET;
	struct isw *isw;
	struct dmraid_dev *rd = NULL;

	if (!(isw = alloc_private(handler, sizeof(*isw))))
		goto out;

	if (!read_dev(di->path, (void *) isw, isw_sboffset, sizeof(*isw))) {
		log_err("%s: failed reading first sector of ISW superblock "
			"for %s", handler, di->path);
		goto bad;
	}

	/* Check Signature */
	to_cpu(isw, 0);
	if (strncmp(isw->sig, MPB_SIGNATURE, sizeof(MPB_SIGNATURE) - 1))
		goto bad;

	/* Check version info, older versions supported */
	if (strncmp(isw->sig + sizeof(MPB_SIGNATURE) - 1,
		   MPB_VERSION_RAID2, sizeof(MPB_VERSION_RAID2) - 1) > 0) {
		log_print("%s: superblock found, but version is too high",
			  handler);
		goto bad;
	}

	/* Get the rounded up value for mpbblocks */
	mpbblocks = ((isw->mpbSize + ISW_DISK_BLOCK_SIZE - 1) /
		    ISW_DISK_BLOCK_SIZE);
	size = mpbblocks * ISW_DISK_BLOCK_SIZE;

	/* Allocate memory for extend Intel superblock */
	dbg_free(isw);
	if (!(isw = alloc_private(handler, size)))
		goto out;

	/* Calculate offset to extended Intel superblock and try to read it */
	isw_sboffset -= (mpbblocks - 1) * ISW_DISK_BLOCK_SIZE;
	if (!read_dev(di->path, (void *) isw, isw_sboffset, size)) {
		log_err("%s: failed reading extended superblock for %s",
			handler, di->path);
		goto bad;
	}

	/* Calculate superblock checksum */
	to_cpu(isw, 1);
	if (isw->checkSum != _checksum(isw)) {
		log_err("%s: extended superblock for %s has wrong checksum",
			handler, di->path);
			goto bad;
	}

	file_metadata(handler, di, isw, sizeof(*isw));

	if (!(rd = alloc_dev()))
		goto bad;

	if (setup_rd(rd, isw, di, isw_sboffset, size))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(isw);

   out:
	return rd;
}

/*
 * Write an Intel Software RAID device.
 */
static int isw_write(struct dmraid_dev *rd, int erase)
{
	int ret;
	struct isw *isw = rd->meta_areas[0].area;

	to_disk(isw, 1);
	ret = write_metadata(handler, rd, -1, erase);
	to_cpu(isw, 1);
	return ret;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an ISW RAID device.
 */
#define	P(f, x, y...) \
	do { \
		log_print("0x%03x " f, \
			  (unsigned int) &x - (unsigned int) isw, y); \
	} while(0)

#define	DP(f, x) do { P(f, x, x);} while(0)

static void isw_log(struct dmraid_dev *rd)
{
	unsigned int d, i;
	struct isw *isw = (struct isw*) rd->meta_areas[0].area;

	log_print("%s (%s):", rd->di->path, handler);
	P("sig: \"%*s\"", isw->sig, MAX_SIGNATURE_LENGTH, isw->sig);
	DP("checkSum: %u", isw->checkSum);
	DP("mpbSize: %u", isw->mpbSize);
	DP("familyNum: %u", isw->familyNum);
	DP("generationNum: %u", isw->generationNum);
	DP("reserved[0]: %u", isw->reserved[0]);
	DP("reserved[1]: %u", isw->reserved[1]);
	DP("numDisks: %u", isw->numDisks);
	DP("numRaidDevs: %u", isw->numRaidDevs);
	DP("fill[0]: %u", isw->fill[0]);
	DP("fill[1]: %u", isw->fill[1]);
	for (i = 0; i < 39; i++) {
		if (isw->filler[i])
        		P("filler[%i]: %u",
				  isw->filler[i], i, isw->filler[i]);
	}

	for (d = 0; d < isw->numDisks; d++) {
		if (!isw->diskTbl[d].totalBlocks)
			continue;

		P("diskTbl[%u].serial: \"%*s\"",
			  isw->diskTbl[d].serial, d,
			  MAX_RAID_SERIAL_LEN, isw->diskTbl[d].serial);
		P("diskTbl[%u].totalBlocks: %u",
			  isw->diskTbl[d].totalBlocks,
			  d, isw->diskTbl[d].totalBlocks);
		P("diskTbl[%u].scsiId: 0x%x",
			  isw->diskTbl[d].scsiId, d, isw->diskTbl[d].scsiId);
		P("diskTbl[%u].status: 0x%x",
			  isw->diskTbl[d].status, d, isw->diskTbl[d].status);
		for (i = 0; i < 5; i++)
			if (isw->diskTbl[d].filler[i])
                		P("diskTbl[%u].filler[%u]: %u",
				  	isw->diskTbl[d].filler[i],
				  	d, i, isw->diskTbl[d].filler[i]);
	}

	for (d = 0; d < isw->numRaidDevs; d++) {
		struct isw_RaidDev* ird = raiddev(isw);

		P("isw_RaidDev[%u].volume: \"%*s\"",
			  ird[d].volume, d, MAX_RAID_SERIAL_LEN, ird[d].volume);
		P("isw_RaidDev[%u].totalSize: %u",
			  ird[d].totalSize, d, ird[d].totalSize);
		P("isw_RaidDev[%u].thisDevSize: %u",
			  ird[d].thisDevSize, d, ird[d].thisDevSize);
		P("isw_RaidDev[%u].thisDevSize256: %u",
			  ird[d].thisDevSize256, d, ird[d].thisDevSize256);
		P("isw_RaidDev[%u].stride: %u",
			  ird[d].stride, d, ird[d].stride);
	}
}
#undef	P
#undef	DP
#endif

/* Add a Intel RAID device to a set */
static struct dmraid_set *isw_group(struct dmraid_dev *dev)
{
	/* FIXME: code grouping logic */
	log_err("%s: no RAID set grouping logic yet", handler);
	return 0;
}

static struct dmraid_format isw_format = {
	.name	= "isw",
	.descr	= "(0) Intel Software RAID",
	.fmt_type = FMT_RAID,
	.read	= isw_read,
	.write	= isw_write,
	.group	= isw_group,
#ifdef DMRAID_NATIVE_LOG
	.log	= isw_log,
#endif
	.type	= isw_type,
};

/* Register this format handler with the format core */
int register_isw(void) {
	register_format(&isw_format);
	return 1;
}

/* Set the RAID device contents up derived from the Intel ones */
static int setup_rd(struct dmraid_dev *rd, struct isw *isw,
		    struct disk_info *di, loff_t isw_sboffset,
		    size_t meta_size)
{
	rd->di = di;
        rd->name = _name(isw);
	rd->fmt = &isw_format;

	rd->offset = ISW_DATAOFFSET;
	rd->sectors = isw_sboffset / ISW_DISK_BLOCK_SIZE; /* FIXME: Correct ? */

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = rd->sectors; /* FIXME: correct ? */
	rd->meta_areas[0].size = meta_size;
	rd->meta_areas[0].area = (void*) isw;

	return 1;
}
