/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * LSI Logic MegaRAID (and MegaIDE ?) ATARAID metadata format handler.
 *
 * Needs more metadata reengineering and grouping logic coding.
 *
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "lsi.h"
#include <utils.h>

static const char *handler = "lsi";

/* Make up RAID device name. */
/* FIXME: senseful name ;) */
static size_t __name(struct lsi *lsi, char *str, size_t len)
{
	return snprintf(str, len, "lsi_%s", "dc") + 1;
}

static char *_name(struct lsi *lsi, int type)
{
	size_t len;
	char *ret;

	len = __name(lsi, NULL, 0);
	if ((ret = dbg_malloc(len + 1))) {
		__name(lsi, ret, len);
		mk_alpha(ret);
	} else
                log_err("%s: Failed allocating name string", handler);

	return ret;
}

/* Mapping of LSI Logic types to generic types */
static struct levels levels[] = {
	{ LSI_MAGIC_T_RAID0, l_raid0 },
	{ LSI_MAGIC_T_RAID1, l_raid1 },
	{ LSI_MAGIC_T_RAID10, l_raid1 },
        { 0, l_undef}
};

/* Neutralize disk type */
static enum level_type lsi_type(struct dmraid_dev *rd) {
	return rd_type(levels,
		       (unsigned int) ((struct lsi*) rd->meta_areas[0].area)->type);
}

/*
 * Read an LSI Logic RAID device
 */
/* Endianess conversion. */
static void to_cpu(struct lsi *lsi)
{
	struct lsi_disks *dsks;

	lsi->stride = le16_to_cpu(lsi->stride);

	for (dsks = lsi->disks; dsks < &lsi->disks[4]; dsks++) {
		dsks->magic_0 = le16_to_cpu(dsks->magic_0);
		dsks->magic_1 = le16_to_cpu(dsks->magic_1);
	}

	lsi->what = le32_to_cpu(lsi->what);
}

static int setup_rd(struct dmraid_dev *rd, struct lsi *lsi,
		     struct disk_info *di);
static struct dmraid_dev *lsi_read(struct disk_info *di) {
	struct lsi *lsi;
	struct dmraid_dev *rd = NULL;

	if (!(lsi = alloc_private(handler, sizeof(*lsi))))
		goto out;

	if (!read_dev(di->path, (void *) lsi, LSI_CONFIGOFFSET, sizeof(*lsi)))
		goto bad;

	to_cpu(lsi);
	if (strncmp(lsi->magic_name, LSI_MAGIC_NAME,
		    sizeof(LSI_MAGIC_NAME) - 1))
		goto bad;

	file_metadata(handler, di, lsi, sizeof(*lsi));

	if (!(rd = alloc_dev()))
		goto bad;

	if (setup_rd(rd, lsi, di))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(lsi);

   out:
	return rd;
}

/*
 * Write a LSI Logic RAID device.
 */
static int lsi_write(struct dmraid_dev *rd, int erase)
{
	int ret;
	struct lsi *lsi = rd->meta_areas[0].area;

	to_disk(lsi);
	ret = write_metadata(handler, rd, -1, erase);
	to_cpu(lsi);
	return ret;
}

/* Add a LSI RAID device to a set */
static struct dmraid_set *lsi_group(struct dmraid_dev *dev) {
	/* FIXME: code grouping logic. */
	log_err("%s: No RAID set grouping logic yet", handler);
	return 0;
}

#ifdef DMRAID_NATIVE_LOG
/* Log native information about an LSI Logic RAID device. */
static void lsi_log(struct dmraid_dev *rd)
{
	unsigned int i;
	struct lsi *lsi = (struct lsi*) rd->meta_areas[0].area;

	log_print("%s (%s):", rd->di->path, handler);
	log_print("magic_name: \"%*s\"",
		  LSI_MAGIC_NAME_LENGTH, lsi->magic_name);
	log_print("seqno: %u", lsi->seqno);
	log_print("type: %u", lsi->type);
	log_print("stride: %u", lsi->stride);
	for (i = 0; i < 4; i++) {
		log_print("disks[%u].unknown: %u", i, lsi->disks[i].unknown);
		log_print("disks[%u].magic_0: 0x%x", i, lsi->disks[i].magic_0);
		log_print("disks[%u].magic_1: 0x%x", i, lsi->disks[i].magic_1);
		log_print("disks[%u].disk_number: %u", i, lsi->disks[i].disk_number);
		log_print("disks[%u].set_number: %u", i, lsi->disks[i].set_number);
	}
	log_print("disk_number: %u", lsi->disk_number);
	log_print("set_number: %u", lsi->set_number);
	log_print("what: %u, 0x%x", lsi->what, lsi->what);
}
#endif

static struct dmraid_format lsi_format = {
	.name	= "lsi",
	.descr	= "(0) LSI Logic MegaRAID",
	.fmt_type = FMT_RAID,
	.read	= lsi_read,
	.write	= lsi_write,
	.group	= lsi_group,
#ifdef DMRAID_NATIVE_LOG
	.log	= lsi_log,
#endif
	.type	= lsi_type,
};

/* Register this format handler with the format core */
int register_lsi(void) {
	register_format(&lsi_format);
	return 1;
}

/* Set the RAID device contents up derived from the LSI ones */
static int setup_rd(struct dmraid_dev *rd, struct lsi *lsi,
		     struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(lsi, 0);
	rd->fmt = &lsi_format;

	rd->offset = LSI_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* FIXME: propper size */
	rd->sectors -= rd->sectors & 1;

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = LSI_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*lsi);
	rd->meta_areas[0].area = (void*) lsi;

	return 1;
}
