/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"
#include "ondisk.h"

/* Format handler registry. */
static LIST_HEAD_INIT(registry);

/*
 * Register a RAID device format handler.
 */
void register_format(struct dmraid_format *fmt)
{
	list_add_tail(&fmt->next, &registry);
}

/* Reset format handler registry iterator. */
void *get_format_list(void)
{
	return (void*) registry.next;
}

/* Get next format handler. */
static void *_get_format(void **iter)
{
	struct list_head *ret = *iter;

	if (ret == &registry)
		return NULL;

	*iter = ret->next;
	return (void *) list_entry(ret, struct dmraid_format, next);
}

void *get_format(void **iter, enum fmt_type type)
{
	void *ret = NULL;

	while((ret = _get_format(iter))) {
		if (type == FMT_ALL ||
		    type == ((struct dmraid_format*) ret)->fmt_type)
			break;
	}

	return ret;
}

/* Return format name */
const char *get_format_name(void *fmt)
{
	return ((struct dmraid_format *) fmt)->name;
}

/*
 * Register all format handlers.
 */
static struct register_fh {
	int (*func)(void);
} register_fh[] = {
#include "register.h"
	{ NULL},
};

int register_format_handlers(void)
{
	int ret = 1;
	struct list_head *elem, *tmp;
	struct register_fh *fh;

	for (fh = register_fh; fh->func; fh++) {
		if (!(ret = fh->func()))
			goto bad;
	}

	return ret;

  bad:
	/* Clean up in case of error */
	log_err("Failed registering format");
	list_for_each_safe(elem, tmp, &registry)
		list_del(&(list_entry(elem, struct dmraid_format, next)->next));

	INIT_LIST_HEAD(&registry);
	return ret;
}

/* Allocate private space in format handlers (eg, for on-disk metadata). */
void *alloc_private(const char *name, size_t size)
{
	void *ret;

	if ((ret = dbg_malloc(size)))
		memset(ret, 0, size);
	else
		log_err("Failed allocating %s metadata", name);

	return ret;
}

/* Allocate metadata sector array in format handlers. */
void *alloc_meta_areas(struct dmraid_dev *rd, const char *name, unsigned int n)
{
	void *ret;
	size_t size = n * sizeof(*rd->meta_areas);

	if ((ret = dbg_malloc(size))) {
		memset(ret, 0, size);
		rd->areas = n;
	} else
		log_err("Failed allocating %s metadata sector array", name);

	return ret;
}

/* Simple metadata write function for format handlers */
static int _write_metadata(const char *handler, struct dmraid_dev *rd,
			      int idx, int erase)
{
	int ret;
	void *p = rd->meta_areas[idx].area, *tmp = p;

	if (erase &&
	    !(p = alloc_private(handler, rd->meta_areas[idx].size)))
		return 0;

	ret = write_dev(rd->di->path, (void *) p,
			rd->meta_areas[idx].offset << 9,
			rd->meta_areas[idx].size);

	if (p != tmp)
		dbg_free(p);

	return ret;
}

int write_metadata(const char *handler, struct dmraid_dev *rd,
		      int idx, int erase)
{
	unsigned int i;

	if (idx > -1)
		return _write_metadata(handler, rd, idx, erase);

	for (i = 0; i < rd->areas; i++)
	{
		if (!_write_metadata(handler, rd, i, erase))
			return 0;
	}

	return 1;
}
