/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * MSDOS partition format handler.
 *
 * msdos_read() and msdos_group influenced by libparted.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "msdos.h"
#include <utils.h>

static const char *handler = "msdos";

/* Make up RAID device name. */
static size_t __name(struct dmraid_dev *rd, unsigned int partition,
		     char *str, size_t len, unsigned char type)
{
	char *p, part[5];

	p = strrchr(rd->di->path, '/');
	p = p ? p + 1 : rd->di->path;
	snprintf(part, 5, "_p%u", partition);
	return snprintf(str, len, "%s%s", p, type ? part : "") + 1;
}

static char *_name(struct dmraid_dev *rd, unsigned int partition,
		   unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(rd, partition, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1)))
		__name(rd, partition, ret, len, type);
	else
		log_err("%s: Failed allocating name string", handler);

	return ret;
}

/* Mapping of msdos type to generic type. */
static enum level_type msdos_type(struct dmraid_dev *rd)
{
	return l_partitions;
}

/*
 * Read a Template RAID device
 */
/* Endianess conversion. */
static void to_cpu(struct msdos *msdos)
{
	struct msdos_partition *part = msdos->partitions;

	for (part = msdos->partitions; part < &msdos->partitions[4]; part++) {
		part->start  = le32_to_cpu(part->start);
		part->length = le32_to_cpu(part->length);
	}

	msdos->magic = le16_to_cpu(msdos->magic);
}

/* Allocate an MSDOS partition sector struct and read the data. */
static int _alloc_and_read(struct msdos **msdos, char *path, uint64_t sector)
{
	struct msdos *m;

	if (!(m = alloc_private(handler, sizeof(*m))))
		return 0;

	if (!read_dev(path, (void *) m, sector << 9, sizeof(*m))) {
		dbg_free(m);
		return 0;
	}

	to_cpu(m);
	*msdos = m;
	return 1;
}

static int setup_rd(struct dmraid_dev *rd, struct msdos *msdos,
		     struct disk_info *di);
static struct dmraid_dev *msdos_read(struct disk_info *di)
{
	struct msdos *msdos;
	struct msdos_partition *p;
	struct dmraid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata */
	if (!_alloc_and_read(&msdos, di->path, MSDOS_CONFIGOFFSET >> 9))
		goto out;

	/* Check magic. */
	if (msdos->magic != MSDOS_MAGIC)
		goto bad;

	for (p = msdos->partitions; p < &msdos->partitions[4]; p++) {
		if (p->type == PARTITION_GPT)
			goto bad;
	}

	if (dump())
		log_print("%s: Filing metadata not supported (use fdisk "
			  "and friends)", handler);

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_dev()))
		goto bad;

	/*
	 * Setup generic RAID device structure members using
	 * ondisk inforation
	 */
	if (setup_rd(rd, msdos, di))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(msdos);

   out:
	return rd;
}

/* Support functions for msdos_group to read the partition table(s). */
static int part_is_extended (struct msdos_partition *part)
{
	switch (part->type) {
	case PARTITION_EXT:
	case PARTITION_EXT_LBA:
	case PARTITION_LINUX_EXT:
		return 1;
	}

	return 0;
}

/* Get a device offset relative to a partition. */
static uint64_t linear_start(const struct msdos_partition *raw_part,
			     uint64_t offset)
{
	return (uint64_t) raw_part->start + offset;
}

/* RAID set allocation support function. */
static struct dmraid_set *_alloc_set(struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	if ((rs = find_set(rd->name, 0)))
		LOG_ERR(NULL, "%s: RAID set %s already exists",
			handler, rs->name);
	else {
		if (!(rs = alloc_set()))
			LOG_ERR(NULL, "%s: Failed allocating RAID set",
				handler);

		rs->level = msdos_type(rd);
		rs->status = s_operational;
	}

	return rs;
}

/*
 * Allocate a RAID device and a set, set the
 * device up and add it to the set.
 */
static int _create_rs(struct dmraid_dev *rd, struct msdos_partition *raw_part,
		      uint64_t sector, unsigned int extended, unsigned int part)
{
	struct dmraid_dev *r;
	struct dmraid_set *rs;

	if (!(r = alloc_dev()))
		return 0;

	if (!(rs = _alloc_set(r))) {
		free_dev(r);
		return 0;
	}

        r->di = rd->di;
	r->fmt = rd->fmt;

        r->name = _name(rd, part, 1);
	r->offset = linear_start(raw_part, sector);
	r->sectors = raw_part->length - 1;

        rs->name = r->name;

	list_add_tail(&r->next, &rs->devs);
	list_add_tail(&rs->next, default_set_list());

	return 1;
}

/*
 * RAID set grouping.
 *
 * In this case it is not really about grouping, more about providing
 * the propper input to the activation layer by creating a RAID set per
 * logical partition and a RAID dev hanging off it mapping the partition
 * linearly.
 *
 * Partition code inspired by libparted and squeezed for this purpose.
 */
static int group_rd(struct dmraid_dev *rd, uint64_t sector, int extended)
{
	unsigned int i, part = extended + 1, ret = 0;
	enum PartitionType type;
	uint64_t part_start;
	struct msdos *msdos = rd->meta_areas[0].area;
	struct msdos_partition *raw_part;

	if (extended && !_alloc_and_read(&msdos, rd->di->path, sector))
		return 0;

	/* Weird: empty extended partitions are filled with 0xf6 by PM */
	if (extended && (msdos->magic == PARTITION_MAGIC_MAGIC))
		return 0;

	/* Walk 4 array entries of a partition table */
	for (i = 0; i < 4; i++, part++) {
		raw_part = &msdos->partitions[i];
		if (raw_part->type == PARTITION_EMPTY || !raw_part->length)
			continue;

		/* FIXME: check integrity of partition's C/H/S values */
		// if (!_check_integrity(raw_part))
		//	goto out;

		part_start = linear_start(raw_part, part > 6 ?
					  DM_SECTORS: sector);
		if (part_start == sector ||
		    part_start > rd->sectors)
			continue;	/* Avoid infinite recursion. */

		type = part_is_extended(raw_part) ? PARTITION_EXTENDED : 0;
		if (type == PARTITION_EXTENDED) {
			if (!group_rd(rd, part_start,
				      extended ? extended + 1 : 4))
				goto out;
		} else if (!_create_rs(rd, raw_part, sector, extended, part))
			goto out;
	}

	ret = 1;

  out:
	if (extended)
		dbg_free(msdos);

	return ret;
}

/* Add a Template RAID device to a set */
static struct dmraid_set *msdos_group(struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	/*
	 * Once we get here, an MSDOS boot sector with a partition table
	 * has been discovered by msdos_read.
	 *
	 * We need to run through the (nested) partiton tables and create
	 * a RAID set and a linear RAID device hanging off it for every primary
	 * and logical partition, so that the activate code is happy.
	 */
	rs = group_rd(rd, 0, 0) ? (struct dmraid_set *) 1 : NULL;

	/* Free the RD. We don't need it any more. */
	free_dev(rd);

	return rs;
}

/*
 * Check integrity of a Highpoint 45X RAID set.
 */
static int msdos_check(struct dmraid_set *rs)
{
	return 1; /* Nice, eh ? */
}

static struct dmraid_format msdos_format = {
	.name	= "msdos",
	.descr	= "(-) MSDOS partitions on SW RAIDs",
	.fmt_type = FMT_PARTITION,
	.read	= msdos_read,
	.group	= msdos_group,
	.check	= msdos_check,
	.type	= msdos_type,
};

/* Register this format handler with the format core */
int register_msdos(void)
{
	register_format(&msdos_format);
	return 1;
}

/*
 * Set the RAID device contents up derived from the MSDOS ones
 *
 * For an MSDOS partition we essentially just save the partition table sector
 * and let msdos_group do the rest...
 */
static int setup_rd(struct dmraid_dev *rd, struct msdos *msdos,
		     struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(rd, 0, 0);
	rd->fmt = &msdos_format;
	rd->offset = 0;
	rd->sectors = di->sectors;

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = MSDOS_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*msdos);
	rd->meta_areas[0].area = (void *) msdos;

	return 1;
}
