/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Template to implement metadata format handlers.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "template.h"
#include <utils.h>

/* Make up RAID device name. */
/* CODEME: implement creation of senseful name for the RAID device */
static size_t __name(struct template *template, char *str, size_t len,
		     unsigned char type)
{
	return snprintf(str, len, "template") + 1;
}

static char *_name(struct template *template, unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(template, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(template, ret, len, type);
		mk_alpha(ret);
	} else
		log_err("Failed allocating name string");

	return ret;
}

/* Mapping of template types to generic types */
/*
 * CODEME: mappings of template private level types to generic ones 
 * (see metadata.h for generic ones)
 */
static struct levels levels[] = {
	{ TEMPLATE_T_SPAN,  l_linear },
	{ TEMPLATE_T_RAID0, l_raid0 },
        { 0, l_undef}
};

/* Neutralize disk type using generic metadata type mapping function */
static enum level_type template_type(struct dmraid_dev *rd)
{
	return rd_type(levels,
		       (unsigned int) ((struct template*) rd->meta_areas[0].area)->type);
}

/*
 * Read a Template RAID device
 */
/* CODEME: enianess conversion */
static void to_cpu(struct template *template)
{
	template->something = le32_to_cpu(template->something);
	...
}

static int setup_rd(struct dmraid_dev *rd, struct template *template,
		     struct disk_info *di);
static struct dmraid_dev *template_read(struct disk_info *di)
{
	struct template *template;
	struct dmraid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata */
	if (!(template = alloc_private("template", sizeof(*template))))
		goto out;

	/* CODEME: calculate TEMPLATE_CONFIGOFFSET correctly */
	if (!read_dev(di->path, (void *) template, TEMPLATE_CONFIGOFFSET,
		      sizeof(*template)))
		goto bad;

	to_cpu(template);

	/* CODEME: unique identification of template metadata */
	if (strncmp(template->magic_name, TEMPLATE_MAGIC_NAME,
		    sizeof(TEMPLATE_MAGIC_NAME) - 1))
		goto bad;
	
	/* File metadata for tests (writes ${devicename}.dat file)
	file_metadata("TEMPLATE", di, template, sizeof(*template));

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_dev()))
		goto bad;

	/*
	 * Setup generic RAID device structure members using
	 * ondisk inforation
	 */
	if (setup_rd(rd, template, di))
		goto out;

	free_dev(rd);
	rd = NULL;

   bad:
	dbg_free(template);

   out:
	return rd;
}

/* Add a Template RAID device to a set */
static struct dmraid_set *template_group(struct dmraid_dev *dev)
{
	/* CODEME: add grouping logic */
	log_print("Implement grouping logic for RAID set");
	return 0;
}

/* CODEME: Write private RAID metadata to device */
static int template_write(struct dmraid_dev *rd, int erase)
{
	int ret;
	struct template *template = rd->meta_areas[0].area;

	/* CODEME: in case there's more complex metadata areas */
	to_disk(template);
	ret = write_metadata(handler, rd, -1, erase);
	to_cpu(template);
	return ret;
}

/*
 * Check integrity of a Highpoint 45X RAID set.
 */
static int template_check(struct dmraid_set *rs)
{
	/* CODEME: implement */
	log_print("Implement checks of RAID set integrity");
	return 0;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about the RAID device.
 */
static void template_log(struct dmraid_dev *rd)
{
	struct template *template = (struct template*) rd->meta_areas[0].area;

	/* CODEME: implement */
	log_print("Implement displaying metadata variables");
}
#endif

static struct dmraid_format template_format = {
	.name	= "template",
	.descr	= "(-) Template RAID",
	.fmt_type = FMT_RAID,
	.read	= template_read,
	.write	= template_write,
	.group	= template_group,
	.check	= template_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= template_log,
#endif
	.type	= template_type,
};

/* Register this format handler with the format core */
int register_template(void)
{
	register_format(&template_format);
	return 1;
}

/* CODEME: Set the RAID device contents up derived from the TEMPLATE ones */
static int setup_rd(struct dmraid_dev *rd, struct template *template,
		     struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(template, 0);
	rd->fmt = &template_format;

	rd->offset = TEMPLATE_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* CODEME: correct sectors */

	if (!(rd->meta_areas = alloc_meta_areas(rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = TEMPLATE_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*template);
	rd->meta_areas[0].area = (void *) template;

	return 1;
}
