/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"

/* Linked list of RAID sets. */
static LIST_HEAD_INIT(raid_sets);

/*
 * Calculate the size of the set by recursively
 * summing the size of the devices in the subsets.
 */
uint64_t calc_total_sectors(struct dmraid_set *rs)
{
        uint64_t sectors = 0;
        struct dmraid_set *r;
        struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, next) {
                sectors += calc_total_sectors(r);
		if (rs->level == l_raid1)
			break;
	}

	list_for_each_entry(rd, &rs->devs, next) {
                sectors += rd->sectors;
		if (rs->level == l_raid1)
			break;
	}

	if (rs->level == l_raid3 || rs->level == l_raid5) {
		sectors -= list_empty(&rs->sets) ?
			   list_entry(rs->devs.next, typeof(*rd), next)->sectors :
			   calc_total_sectors(list_entry(rs->sets.next, typeof(*rs), next));
	} 

        return sectors;
}

/* Count devices in a set. */
unsigned int count_devs(struct dmraid_set *rs)
{
	unsigned int ret = 0;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, next)
		ret += count_devs(r);

	list_for_each_entry(rd, &rs->devs, next)
		ret++;

	return ret;
}

/* Allocate/Free RAID device (member of a RAID set). */
struct dmraid_dev *alloc_dev(void)
{
	struct dmraid_dev *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret));
		INIT_LIST_HEAD(&ret->next);
	} else
		log_err("Failed allocating RAID device");

	return ret;
}

void free_dev(struct dmraid_dev *rd)
{
	if (rd->name)
		dbg_free(rd->name);

	if (rd->meta_areas)
		dbg_free(rd->meta_areas);

	dbg_free(rd);
}

/* Reference to global default set. */
struct list_head *default_set_list(void)
{
	return &raid_sets;
}

/* Allocate/Free RAID set. */
struct dmraid_set *alloc_set(void)
{
	struct dmraid_set *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret)); /* Clear needed! */
		INIT_LIST_HEAD(&ret->next);
		INIT_LIST_HEAD(&ret->sets);
		INIT_LIST_HEAD(&ret->devs);
	} else
		log_err("Failed allocating RAID set");

	return ret;
}

void free_set(struct dmraid_set *rs)
{
	dbg_free(rs);
}


/* Remove a set recursively. */
void remove_set(struct dmraid_set *rs)
{
	struct list_head *elem, *tmp;

	list_for_each_safe(elem, tmp, &rs->sets)
		remove_set(list_entry(elem, struct dmraid_set, next));

	list_for_each_safe(elem, tmp, &rs->devs) {
		list_del(elem);
		free_dev(list_entry(elem, struct dmraid_dev, next));
	}

	list_del(&rs->next);
	free_set(rs);
}

/* Free all sets registered */
void free_all_sets(void)
{
	struct list_head *elem, *tmp;

	list_for_each_safe(elem, tmp, default_set_list())
		log_print("%s", (list_entry(elem, struct dmraid_set, next))->name);

	list_for_each_safe(elem, tmp, default_set_list())
		remove_set(list_entry(elem, struct dmraid_set, next));

	if (!list_empty(default_set_list()))
		log_err("FATAL: default_set_list() not empty");
}

/* Add a device to the end of the list */
void list_add_dev_sorted(struct dmraid_set *rs, struct dmraid_dev *new,
			 int (*f)(struct dmraid_dev *rd,
				  struct dmraid_dev *new))
{
	struct dmraid_dev *rd;

	list_for_each_entry(rd, &rs->devs, next) {
		/*
		 * Add in at the beginning or where the metadata
		 * format handler function tells us to.
		 */
		if (!f || f(rd, new)) {
			list_add(&new->next, &rd->next);
			return;
		}
	}

	list_add_tail(&new->next, &rs->devs);
}

/* Return a RAID set list iterator. */
void *start_set_list(void)
{
	return (void*) (default_set_list())->next;
}

/* Get next set off the list. */
void *get_set(void **iter)
{
	struct list_head *ret = *iter;

	if (ret == &raid_sets)
		ret = NULL;
	else
		*iter = ret->next;

	return ret ? (void *) list_entry(ret, struct dmraid_set, next) : NULL;
}

/* Return RAID set name. */
const char *get_set_name(void *rs)
{
	return ((struct dmraid_set *) rs)->name;
}

/* Return != 0 in case of a partitioned RAID set. */
int partitioned_set(void *rs)
{
	return ((struct dmraid_set*) rs)->level == l_partitions;
}
/* Return != 0 in case of a partitioned RAID set. */
int base_partitioned_set(void *rs)
{
	return ((struct dmraid_set*) rs)->has_partitions;
}

/* Find RAID set by name (recursive if decend != 0). */
static struct dmraid_set *_find_set(struct list_head *list,
				    const char *name, int decend) {
	struct dmraid_set *rs;

	if (list_empty(list) || !name)
		return NULL;

	log_dbg(1, "_find_set: searching %s", name);

	list_for_each_entry(rs, list, next) {
		if (!strcmp(rs->name, name)) {
			log_dbg(1, "_find_set: found %s", name);
			return rs;
		}

		if (decend && !list_empty(&rs->sets))
			return _find_set(&rs->sets, name, decend);
	}

	return NULL;
}

struct dmraid_set *find_set(const char *name, int decend)
{
	return _find_set(&raid_sets, name, decend);
}

/* Return # of raid sets build */
unsigned int count_sets(void)
{
	int ret = 0;
	struct list_head *elem;

	list_for_each(elem, default_set_list())
		ret++;

	return ret;
}

/*
 * Support function for format handlers:
 *
 * Return neutralized RAID type for given mapping array (linear, raid0, ...)
 */
enum level_type rd_type(struct levels* l, unsigned int level)
{
	for (; l->type != l_undef; l++) {
		if (level == l->level)
			return l->type;
	}

	return l_undef;
}

/*
 * Read RAID metadata off a device by trying
 * all registered format handlers in turn.
 */
static struct dmraid_dev *_dmraid_read(struct disk_info *di,
				       struct dmraid_format *fmt)
{
	struct dmraid_dev *rd = NULL;

	log_verbose(2, "%s: %-6s discovering", di->path, fmt->name);

	if ((rd = fmt->read(di))) {
		log_verbose(2, "%s: %s metadata discovered",
			    di->path, fmt->name);
		rd->fmt = fmt;
	}

	return rd;
}

static struct dmraid_dev *dmraid_read(struct disk_info *di, char const *format,
				      int partitions)
{
	void *iter;
	struct dmraid_format *fmt;
	struct dmraid_dev *rd = NULL, *rd_tmp = NULL;

	for (iter = get_format_list(); (fmt = get_format(&iter, FMT_ALL)); ) {
		if (fmt->fmt_type == FMT_PARTITION) {
			if (!partitions)
				continue;
		} else if (partitions)
			continue;
		else if (format && strcmp(format, fmt->name))
			continue;

		if ((rd_tmp = _dmraid_read(di, fmt))) {
			if (rd) {
				log_err("%s: WARNING, \"%s\" and \"%s\" "
					"formats discovered "
					"(using %s)!",
					di->path, rd_tmp->fmt->name,
					rd->fmt->name, rd->fmt->name);
				free_dev(rd_tmp);
			} else
				rd = rd_tmp;
		}
	}

	if (rd_tmp && rd_tmp != rd) {
		free_dev(rd);
		rd = NULL;
	}

	return rd;
}

/*
 * Write RAID metadata to a device.
 */
static int dmraid_write(struct dmraid_dev *rd, int erase)
{
	int ret = 0;
	struct dmraid_format *fmt = rd->fmt;

	if (fmt->write) {
		log_verbose(1, "%s %s",
			    erase ? "Erasing metadata on" :
				    "Writing metadata to %s",
			    rd->di->path);
		ret = fmt->write(rd, erase);
	} else
		log_err("Format \"%s\" doesn't support writing metadata",
			fmt->name);

	return ret;
}

/*
 * Group RAID device into a RAID set.
 */
static struct dmraid_set *dmraid_group(struct dmraid_dev *rd)
{
	return rd->fmt->group(rd);
}

/* Check that device names are members of devices list. */
static int _device_hit(struct disk_info *di, char **devices)
{
	char **dev;

	if (!devices || !*devices)
		return 1;

	for (dev = devices; *dev; dev++) {
		if (!strcmp(*dev, di->path))
			return 1;
	}

	return 0;
}

/*
 * Discover RAID devices.
 */
void discover_rd(char *fmt, char **devices)
{
	void *iter;
	struct disk_info *di;
	struct dmraid_dev *rd;

	for (iter = rewind_disk_info(); (di = get_disk_info(&iter)); ) {
		if (_device_hit(di, devices) &&
		    (rd = dmraid_read(di, fmt, 0)))
			di->rd = rd;
	}
}

/*
 * Discover partitions on RAID sets.
 */
void discover_partitions(void)
{
	char *path;
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs, *r;

  redo:
	list_for_each_entry(rs, default_set_list(), next) {
		if (rs->has_partitions ||
		    partitioned_set(rs) ||
		    !dm_status(rs))
			continue;

		if (!(path = mkdm_path(rs->name)))
			return;

		if (!(di = alloc_disk_info(path)))
			goto bad;

		di->sectors = calc_total_sectors(rs);
		if (!(rd = dmraid_read(di, NULL, 1))) {
			free_disk_info(di);
			continue;
		}

		di->rd = rd;

		/*
		 * WARNING: partition group function returns
		 * a dummy pointer we don't want to access!
		 */
		if ((r = dmraid_group(rd))) {
			log_verbose(1, "Added %s to partition RAID set", path);
			rs->has_partitions = 1;
			goto redo;
		} else {
			log_err("Failed adding %s to RAID set", path);
			free_disk_info(di);
		}
	}
	return;

  bad:
	dbg_free(path);
}

/*
 * Count devices found
 */
unsigned int count_devices(enum dev_type type)
{
	unsigned int ret = 0;
	void *iter;
	struct disk_info *di;

	for (iter = rewind_disk_info(); (di = get_disk_info(&iter));) {
		if (type == DEVICE)
			ret++;
		else if (type & (RAID|NATIVE) && di->rd)
			ret++;
	}

	return ret;
}

/* Get format handler of RAID set */
static struct dmraid_format *_get_fmt(struct dmraid_set *rs)
{
	struct dmraid_dev *rd;

	while (!list_empty(&rs->sets))
		rs = list_entry(rs->sets.next, typeof(*rs), next);

	rd = list_entry(rs->devs.next, typeof(*rd), next);
	return rd ? rd->fmt : NULL;
}

/*
 * Check metadata consistency of raid sets.
 */
static void check_sets(void)
{
	struct list_head *elem, *tmp;
	struct dmraid_set *rs;
	struct dmraid_format *fmt;

	list_for_each_safe(elem, tmp, &raid_sets) {
		rs = list_entry(elem, typeof(*rs), next);
		if (!(fmt = _get_fmt(rs)))
			continue;

		if (!fmt->check) {
			log_err("Fatal: Format \"%s\" lags a check method!",
				fmt->name);
			remove_set(rs);
			continue;
		}

		if (!fmt->check(rs)) {
			log_err("RAID set \"%s\" is inconsistent", rs->name);
			/*
			 * FIXME: check needed if degraded activation
			 *        is sensible.
			 */
			if (rs->level == l_raid1)
				log_err("Check of mirrored set \"%s\" failed, "
					"keeping it", rs->name);
			else {
				log_err("Check of \"%s\" failed, removing set",
					rs->name);
				remove_set(rs);
			}
		}
	}

	return;
}

/*
 * Build RAID set(s)
 *
 *	name = NULL  : build all sets
 *	name = String: build just the one set
 *
 *	format = NULL  : try discovering RAID devices using
 *                       all registered formats
 *	format = String: try discovering RAID devices using the
 *                       specified format only
 *
 */
int build_set(char *name)
{
	void *iter;
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs;

	if (find_set(name, 0))
		LOG_ERR(0, "RAID set %s already exists\n", name);

	for (iter = rewind_disk_info(); (di = get_disk_info(&iter)); ) {
		if (!(rd = di->rd) ||
		    (name && strncmp(rd->name, name, strlen(name))))
			continue;
	
		if ((rs = dmraid_group(rd)))
			log_verbose(1, "Added %s to RAID set %s",
				    rd->di->path, rs->name);
		else {
			if (rd->sectors)
				log_err("Failed adding %s to RAID set",
					rd->di->path);
			else
				log_print("Skipping spare %s", rd->di->path);

			if ((rs = find_set(rd->name, 0))) {
				log_err("Removing set");
				remove_set(rs);
			}
		}
	}

	check_sets();

	return 1;
}

/*
 * Write RAID set metadata to devices.
 */
int write_set(void *v)
{
	int ret = 0;
	struct dmraid_set *r, *rs = v;
	struct dmraid_dev *rd;

	/* Decend hierarchy */
	list_for_each_entry(r, &rs->sets, next) {
		if (!(ret = write_set((void *) r)))
			break;
	}

	/* Write metadata to devices */
	list_for_each_entry(rd, &rs->devs, next) {
		if (!(ret = dmraid_write(rd, 0)))
			break;
	}

	return ret;
}

/*
 * Support function for format handlers:
 *
 * File RAID metadata for analysis.
 */
static size_t _name(char *str, size_t len, char *path, const char *handler)
{
	char *s;
	
	s = strrchr(path, '/');
	s = s ? s + 1 : path;
	return snprintf(str, len, "%s_%s.dat", s, handler) + 1;
}

void file_metadata(const char *handler, struct disk_info *di,
		   void *data, size_t size)
{
	if (dump()) {
		int fd;
		size_t len;
		char *name = NULL;
	
		len = _name(NULL, 0, di->path, handler) + 1;
		if (!(name = dbg_malloc(len))) {
			log_err("Error allocation metadata file memory");
			return;
		}

		_name(name, len, di->path, handler);
		if ((fd = open(name, O_CREAT|O_TRUNC|O_WRONLY, 0600)) == -1) {
			log_err("Error opening metadata file");
			goto free;
		}

		log_verbose(1, "WRITING %s", name);
		if (write(fd, data, size) != size) {
			log_err("Error writing metadata file");
			goto free;
		}
	
		close(fd);
		return;

  free:
		dbg_free(name);
		close(fd);
	}
}

/*
 * Erase ondisk metadata.
 */
int erase_metadata(void)
{
	void *iter;
	struct disk_info *di;

	for (iter = rewind_disk_info(); (di = get_disk_info(&iter)); ) {
		if (!di->rd)
			continue;

		if (yes_no_prompt("Do you really want to erase \"%s\" ondisk "
				  "metadata on %s",
				  di->rd->fmt->name, di->path) &&
		    !dmraid_write(di->rd, 1))
			LOG_ERR(0, "Error erasing ondisk metadata on %s",
				di->path);
	}

	return 1;
}
