/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#ifndef	_METADATA_H_
#define	_METADATA_H_

#include "datastruct/list.h"
#include <stdint.h>

/* Unified RAID levels. */
enum level_type {
	l_undef	= 0,
	l_partitions,
	l_spare,
	l_linear,
	l_raid0,
	l_raid1,
	/* Higher RAID levels below not supported (yet). */
	l_raid3,
	l_raid4,
	l_raid5,
};

/*
 * RAID level struct for RAID level type unification.
 *
 * Format handler allocates an array and inserts mappings
 * between format specific types and the unified ones.
 */
struct levels {
	unsigned int    level; /* Must be long enough for vendor definition. */
	enum level_type type;
};

/* RAID disk status. */
enum status_type {
	s_undef = 0,
	s_setup,
	s_operational,
	s_broken,
	s_nosync,
	s_inconsistent,
};

/* Metadata areas and size stored on a RAID device. */
struct meta_areas {
	uint64_t	offset;	/* on disk metadata offset in sectors */
	size_t		size;	/* on disk metadata size */
	void		*area;	/* pointer to format specific metadata */
};

/*
 * Abstracted RAID device.
 *
 * A RAID device is a member of a RAID set and can only
 * exist at the lowest level of a RAID set stack (eg, for RAID10).
 */
struct dmraid_dev {
	struct list_head next;		/* chain of devices hanging off a set */
	struct disk_info  *di;		/* backpointer to disk_info */

	char		*name;		/* Name of set this device belongs to */

	struct dmraid_format *fmt;	/* Format handler for this device */

	uint64_t	offset;		/* Data offset on device */
	uint64_t	sectors;	/* Length of the segment to map */

	unsigned int	areas;		/* # of metadata areas on the device */
	struct meta_areas *meta_areas;	/* Dynamic array of metadata areas */

	void		*private;	/* For format handler use (eg, to
					   keep references between calls) */
};

/*
 * Abstracted (hierarchical) RAID set.
 *
 * Can be used to form a tree of subsets with arbitrary depths.
 * Defines RAID attributes for the set as a whole (eg: RAID0, Status).
 */
struct dmraid_set {
	struct list_head next;		/* Chain of independent sets. */
	const char *name;		/* Name of the set. */

	/*
	 * Pointer to chain of sub sets (eg, RAID10) which
	 * make up device-mapper stacks.
	 * Lowest level identified with NULL here.
	 */
	struct list_head sets;

	/*
	 * Pointer to the disks of the RAID volume.
	 *
	 * High sets in the stack will have NULL here.
	 *
	 * Lowest level will (typically) hold devices definitions for
	 * IDE/SCSI/... devices.
	 */
	struct list_head devs;

	unsigned int stride;		/* stripesize */
	enum level_type level;		/* Unified raid level */
	unsigned int has_partitions;	/* Flag to indicate RAID set	
					   has partitions */
	enum status_type status;	/* Status of set */

};

uint64_t calc_total_sectors(struct dmraid_set *rs);
struct dmraid_dev *alloc_dev(void);
void free_dev(struct dmraid_dev *rd);
void list_add_dev_sorted(struct dmraid_set *rs, struct dmraid_dev *new,
			 int (*f)(struct dmraid_dev *rd,
			 struct dmraid_dev *new));
struct dmraid_dev *list_del_dev(struct dmraid_set *rs);
struct dmraid_set *alloc_set(void);
unsigned int count_devs(struct dmraid_set *rs);
void free_set(struct dmraid_set *rs);
void free_all_sets(void);
struct list_head *default_set_list(void);
struct dmraid_set *find_set(const char *name, int decend);
void remove_set(struct dmraid_set *rs);
void *start_set_list(void);
void *get_set(void **iter);
const char *get_set_name(void *rs);
int build_set(char *name);
unsigned int count_sets(void);
int write_set(void *rs);
int partitioned_set(void *rs);
int base_partitioned_set(void *rs);
void discover_rd(char *fmt, char **devices);
void discover_partitions(void);
unsigned int count_devices(enum dev_type type);
enum level_type rd_type(struct levels *levels, unsigned int level);
void file_metadata(const char *handler, struct disk_info *di,
		   void *data, size_t size);
int erase_metadata(void);

#endif
