/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this dmraid code; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include <getopt.h>
#include <string.h>
#include <dmraid.h>
#include "log/log.h"
#include "commands.h"
#include "toollib.h"
#include "version.h"

/* Action flags */
enum action action = UNDEF;

/* Metadata format to use. */
char *format = NULL;

/* Command line options */
static char const *short_opts = "a:p"
#ifndef	DMRAID_MINI
				"bcdDEf:hl"
#ifdef	DMRAID_NATIVE_LOG
				"n"
#endif
				"rs::tvV"
#endif
;

static struct option long_opts[] = {
	{"activate", required_argument, NULL, 'a'},
	{"format", required_argument, NULL, 'f'},
	{"partitions", no_argument, NULL, 'p'},
#ifndef	DMRAID_MINI
	{"block_devices", no_argument, NULL, 'b'},
	{"display_columns", no_argument, NULL, 'c'},
	{"debug", no_argument, NULL, 'd'},
	{"dump_metadata", no_argument, NULL, 'D'},
	{"erase_metadata", no_argument, NULL, 'E'},
	{"help", no_argument, NULL, 'h'},
	{"list_formats", no_argument, NULL, 'l'},
#  ifdef	DMRAID_NATIVE_LOG
	{"native_log", no_argument, NULL, 'n'},
#  endif
	{"raid_devices", no_argument, NULL, 'r'},
	{"sets", optional_argument, NULL, 's'},
	{"test", no_argument, NULL, 't'},
	{"verbose", no_argument, NULL, 'v'},
	{"version", no_argument, NULL, 'V'},
#endif
	{NULL, no_argument, NULL, 0}
};

/* Check activate/deactivate option arguments. */
static int check_activate(void)
{
	size_t len;

	if (!optarg)
		return 0;

	str_tolower(optarg);
	len = strlen(optarg);
	if (!strncmp(optarg, "yes", len))
		action |= ACTIVATE;
	else if (!strncmp(optarg, "no", len))
		action |= DEACTIVATE;
	else
		goto err;

	return 1;

  err:
	LOG_ERR(0, "Invalid option argument for -a");
}

/* Check format identifier */
static int check_format(void)
{
	if (!valid_format(optarg)) {
		log_err("Invalid format \"%s\" for -f (see -l)", optarg);
		return 0;
	}

	return (format = strdup(optarg)) ? 1 : 0;
}

#ifndef	DMRAID_MINI
/* Check active/inactive option arguments. */
static int check_active(void)
{
	size_t len;

	if (!optarg)
		return 1;

	str_tolower(optarg);
	len = strlen(optarg);
	if (!strncmp(optarg, "active", len))
		action |= ACTIVE;
	else if (!strncmp(optarg, "inactive", len))
		action |= INACTIVE;
	else
		goto err;

	return 1;

  err:
	LOG_ERR(0, "Invalid option argument for -s");
}

/* Display help information */
#ifndef	DMRAID_MINI
static int help_run = 0;
static int help(void)
{
	char *c = cmd();

	log_print("%s: Device-Mapper Software RAID tool\n", c);
	log_print("* = [-d|--debug]{0,} [-v|--verbose]{0,}\n");
	log_print("%s\t{-a|--activate} {y[es]/n[o]} *\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[-p|--no_partitions]\n"
		  "\t[-t|--test]\n"
		  "\t[RAID set]{0,}\n", c);
	log_print("%s\t{-b|--block_devices} *\n"
		  "\t[-c|--display_columns]\n", c);
	log_print("%s\t{-h|--help}\n", c);
	log_print("%s\t{-l|--list_formats} *\n", c);
#  ifdef	DMRAID_NATIVE_LOG
	log_print("%s\t{-n|--native_log} *\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[device-path]{0,}\n", c);
#  endif
	log_print("%s\t{-r|--raid_devices} *\n"
		  "\t[-c|--display_columns]\n"
		  "\t[-D|--dump_metadata]\n"
		  "\t[-E|--erase_metadata]\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[device-path]{0,}\n", c);
	log_print("%s\t{-s|--sets}[a[active]/i[nactive]] *\n"
		  "\t[-c|--display_columns]\n"
		  "\t[-f|--format FORMAT]\n"
		  "\t[RAID set]{0,}\n", c);
	log_print("%s\t{-V/--version}\n", c);

	help_run++;
	return 1;
}
#endif

static int version(void)
{
	char v[80] = "\0";

	log_print("%s version:\t\t%s", cmd(), DMRAID_VERSION);
	log_print("%s library version:\t%s", cmd(), libdmraid_version());
	dm_version(v, sizeof(v));
	log_print("device-mapper version:\t%s", v);
	return 1;
}
#endif

/*
 * Action flag definitions for set_action()
 *
 * 'Early' options can be handled directly in set_action() by calling
 * the functions registered here (set on_set member).
 */
static struct actions actions [] = {
	{'a', 1, 1, 0, 1, check_activate, ACTIVATE|DEACTIVATE|FORMAT|NO_PARTITIONS
#ifndef DMRAID_MINI
	|DBG|TEST|VERBOSE
#endif
	},
	{'f', 1, 0, FORMAT, 1, check_format, ACTIVATE|DEACTIVATE
#ifndef DMRAID_MINI
	|ACTIVE|INACTIVE|COLUMN|DBG|DUMP|ERASE
#  ifdef DMRAID_NATIVE_LOG
	|NATIVE_LOG
#  endif
	|RAID_DEVICES|SETS|TEST|VERBOSE
#endif
	},
	{'p', 0, 0, NO_PARTITIONS, 0, NULL, ACTIVATE|DEACTIVATE|FORMAT
#ifndef DMRAID_MINI
	|DBG|NO_PARTITIONS|TEST|VERBOSE
#endif
	},
#ifndef	DMRAID_MINI
	{'b', 0, 1, BLOCK_DEVICES, 0, NULL, COLUMN|DBG|VERBOSE},
	{'c', 1, 0, COLUMN, 1, inc_column, ACTIVE|INACTIVE|BLOCK_DEVICES|DBG|FORMAT|RAID_DEVICES|SETS},
	{'d', 1, 0, DBG, 1, inc_debug, ACTIVATE|DEACTIVATE|BLOCK_DEVICES|DUMP|ERASE|FORMAT|LIST_FORMATS|SETS|NO_PARTITIONS|RAID_DEVICES|TEST|VERBOSE},
	{'D', 1, 0, DUMP, 1, inc_dump, DBG|FORMAT|RAID_DEVICES|VERBOSE},
	{'E', 1, 0, ERASE, 0, NULL, DBG|FORMAT|RAID_DEVICES|VERBOSE},
	{'h', 1, 1, HELP, 0, help, 0},
	{'l', 0, 1, LIST_FORMATS, 0, list_formats, DBG|VERBOSE},
#  ifdef DMRAID_NATIVE_LOG
	{'n', 1, 1, NATIVE_LOG, 0, NULL, DBG|FORMAT|VERBOSE},
#  endif
	{'r', 1, 1, RAID_DEVICES, 0, NULL, COLUMN|DBG|DUMP|ERASE|FORMAT|VERBOSE},
	{'s', 1, 1, SETS, 1, check_active, ACTIVE|INACTIVE|COLUMN|DBG|FORMAT|VERBOSE},
	{'t', 1, 0, TEST, 1, inc_test, ACTIVATE|DEACTIVATE|DBG|FORMAT|NO_PARTITIONS|VERBOSE},
	{'v', 1, 0, VERBOSE, 1, inc_verbose, ACTIVATE|DEACTIVATE|DBG|DUMP|ERASE|BLOCK_DEVICES|FORMAT|LIST_FORMATS|SETS|NO_PARTITIONS|RAID_DEVICES|TEST},
	{'V', 0, 1, VERSION, 0, version, 0},
#endif
	{0, 0, 0, UNDEF, 0, NULL, UNDEF},
};

/*
 * Set action flag and optionally call function on set.
 */
static int set_action(int o)
{
	struct actions *a;

	for (a = actions; a->option; a++) {
		if (o == a->option) {
			action |= a->action;
			a->allowed |= a->action;
			if (a->on_set && !a->func())
				return 0;
		}
	}

	return 1;
}

/* Check for invalid option combinations */
static int check_actions(char **argv, int *end)
{
	struct actions *a;

	for (a = actions; a->option; a++) {
		if (a->action & action) {
			if (!a->alone &&
			    a->action == action)
				LOG_ERR(0, "option -%c can't stand alone\n",
					a->option);

			if (~a->allowed & action)
#ifndef DMRAID_MINI
				LOG_ERR(0, "Invalid option combination%s\n",
					help_run ? "" : " (-h for help)");
#else
				LOG_ERR(0, "Invalid option combination\n");
#endif

			if (!a->args && argv[optind])
				LOG_ERR(0, "No arguments allowed with -%c\n",
					a->option);

			if (!a->on_set &&
			    a->func) {
				(*end)++;
				if (!a->func())
					return 0;
			}
		}
	}

	if (!action)
		LOG_ERR(0, "Options missing\n");

#ifndef DMRAID_MINI
	if ((action & (DBG|VERBOSE)) == action)
		LOG_ERR(0, "Give more options with -d/-v\n");
#endif

	return 1;
}

/* Parse and handle the command line arguments */
int handle_args(int argc, char ***argv, int *end)
{
	int o, opt_idx, ret = 0;

	if (argc < 2)
#ifndef DMRAID_MINI
		LOG_ERR(0, "No arguments/options given (-h for help)\n");
#else
		LOG_ERR(0, "No arguments/options given\n");
#endif

	/* Walk the options (and option arguments) */
	while ((o = getopt_long(argc, *argv, short_opts,
				long_opts, &opt_idx)) != -1) {
		if (o == ':' || o == '?' || !set_action(o))
			return 0;
	}

	/* Force deactivation of stacked partition devices. */
	if (DEACTIVATE & action)
		action &= ~NO_PARTITIONS;

	ret = check_actions(*argv, end);
	*argv = *argv + optind;

	return ret;
}

/*********************************************************************
 * Perform pre/post functions for requested actions.
 */

/* Post Activate/Deactivate RAID set. */
static int _activate(void *rs, int arg)
{
	if ((ACTIVATE & action) ? activate_set(rs) :
				  deactivate_set(rs))
		log_verbose(1, "%sctivated %s RAID set %s",
			    action & ACTIVATE ? "A": "Dea", get_set_level(rs),
			    get_set_name(rs));
	else if (ACTIVATE & action)
		LOG_ERR(0, "Failed %s" "activating %s RAID set",
			action & ACTIVATE ? "" : "de", get_set_level(rs));

	return 1;
}

#ifndef DMRAID_MINI
/* Pre and post display_set() functions. */
static int _display_set_arg(void *v)
{
	return (action & ACTIVE) ? A : ((action & INACTIVE) ? I : ALL);
}

static int _display_set(void *rs, int type)
{
	display_set(rs, type);
	return 1;
}

/* Pre and post display_devices functions. */
static int _devices(void *v)
{
	return DEVICE;
}

#  ifdef DMRAID_NATIVE_LOG
static int _native(void *v)
{
	return NATIVE;
}
#  endif

static int _raid(void *v)
{
	return RAID;
}

static int _display_devices(void *v, int type)
{
	display_devices(type);
	return 1;
}

static int _erase(void *dummy, int arg)
{
	return erase_metadata();
}
#endif

/* Definition of pre- and post functions to perform. */
struct prepost prepost[] = {
	{ ACTIVATE|DEACTIVATE, NULL, _activate },
#ifndef DMRAID_MINI
	{ BLOCK_DEVICES, _devices, _display_devices },
	{ ERASE, NULL, _erase },
	{ SETS, _display_set_arg, _display_set },
#  ifdef DMRAID_NATIVE_LOG
	{ NATIVE_LOG, _native, _display_devices },
#  endif
	{ RAID_DEVICES, _raid, _display_devices },
#endif
	{ 0, NULL, NULL },
};

/*
 * Function abstraction which takes pre- and post-function calls
 * to prepare an argument in pre to be used by post.
 *
 * perform() is the call handler for all functions using already
 * available device and RAID metadata such as displaying, erasing and
 * activation/deactivation of RAID sets.
 *
 * The necessary (meta)data about disks, RAID devices and RAID sets must
 * already be present on calls to this handler!
 */

int perform(void *rs)
{
	int ret = 0;
	struct prepost *p;

	for (p = prepost; p->action; p++) {
		if (p->action & action) {
			if (p->pre)
				ret = p->pre(rs);

			ret = p->post(rs, ret);
			break;
		}
	}

	return ret;
}
